* [Introduction](/)
{%- for package in packages %}
* [{{package}} package]({{package}}.md)
{%- endfor %}
{% if deprecated_plugins %}
* [Deprecated plugins](plugin_deprecations.md)
{% endif %}
{% if deprecated_arguments %}
* [Deprecated arguments](argument_deprecations.md)
{% endif %}
